import { State, UpdateRect } from "./model";

export function createCanvas(parent: HTMLElement) {
  const canvas = document.createElement("canvas");
  canvas.style.cssText = `
    position: fixed;
    left: 0;
    top: 0;
    pointer-events: none;
    z-index: 10000000;
    max-width: none !important;
  `;
  canvas.setAttribute("data-render-highlight", "true");

  parent.appendChild(canvas);

  return canvas;
}

/**
 * Draw on the canvas
 * @returns The interval when the next draw should happen. A return
 * value of -1 means that we can abort rendering
 */
export function draw(
  state: State,
  options: { colors: string[]; outlineColor: string }
): number {
  const { canvas, updates, timer } = state;
  if (!canvas || !canvas.getContext) return -1;
  if (timer) clearTimeout(timer);

  const ctx = canvas.getContext("2d");
  if (!ctx) return -1;
  ctx.clearRect(0, 0, canvas.width, canvas.height);

  const now = performance.now();
  let nextRedraw = Number.MAX_SAFE_INTEGER;
  updates.forEach((data, key) => {
    if (data.expirationTime < now) {
      updates.delete(key);
    } else {
      drawRect(ctx, data, options);
      nextRedraw = Math.min(nextRedraw, data.expirationTime);
    }
  });

  if (nextRedraw !== Number.MAX_SAFE_INTEGER) {
    return nextRedraw - now;
  }

  return -1;
}

function drawRect(
  ctx: CanvasRenderingContext2D,
  data: UpdateRect,
  options: { colors: string[]; outlineColor: string }
) {
  const { colors, outlineColor } = options;
  const colorIndex = Math.min(colors.length - 1, data.count - 1);

  // Outline
  ctx.lineWidth = 1;
  ctx.strokeStyle = outlineColor;
  ctx.strokeRect(data.x - 1, data.y - 1, data.width + 2, data.height + 2);

  // Inset
  ctx.lineWidth = 1;
  ctx.strokeStyle = outlineColor;
  ctx.strokeRect(data.x + 1, data.y + 1, data.width - 2, data.height - 2);

  // Border
  ctx.strokeStyle = colors[colorIndex];
  ctx.lineWidth = 1;
  ctx.strokeRect(data.x, data.y, data.width, data.height);
}
